"""
These settings are here to use during tests, because django requires them.

In a real-world use case, apps in this project are installed into other
Django applications, so these settings will not be used.
"""


from os.path import abspath, dirname, join

from enterprise_data_roles.constants import (
    ENTERPRISE_DATA_ADMIN_ROLE,
    SYSTEM_ENTERPRISE_ADMIN_ROLE,
    SYSTEM_ENTERPRISE_OPERATOR_ROLE,
)


def here(*args):
    """
    Return the absolute path to a directory from this file.
    """
    return join(abspath(dirname(__file__)), *args)


def root(*args):
    """
    Return the absolute path to some file from the project's root.
    """
    return abspath(join(abspath(here('../..')), *args))


DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": "default.db",
        "USER": "",
        "PASSWORD": "",
        "HOST": "",
        "PORT": "",
    }
}

INSTALLED_APPS = (
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sites",
    "django.contrib.sessions",
    "django.contrib.admin",  # only used in DEBUG mode
    "django.contrib.messages",
    "django.contrib.staticfiles",

    "waffle",

    "enterprise_data",
    "enterprise_reporting",
    "enterprise_data_roles",
    "rules.apps.AutodiscoverRulesConfig",

    # drf-jwt
    "rest_framework_jwt",
)

MIDDLEWARE = [
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "crum.CurrentRequestUserMiddleware",
    "waffle.middleware.WaffleMiddleware",
]

AUTHENTICATION_BACKENDS = [
    "rules.permissions.ObjectPermissionBackend",
    "django.contrib.auth.backends.ModelBackend"
]

SESSION_ENGINE = "django.contrib.sessions.backends.file"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": {
                "django.contrib.auth.context_processors.auth",  # this is required for admin
                "django.contrib.messages.context_processors.messages",
            }
        }
    },
]

LMS_BASE_URL = "http://localhost:8000/"

BACKEND_SERVICE_EDX_OAUTH2_KEY = "analytics_api-backend-service-key"
BACKEND_SERVICE_EDX_OAUTH2_SECRET = "analytics_api-backend-service-secret"
BACKEND_SERVICE_EDX_OAUTH2_PROVIDER_URL = "http://127.0.0.1:8000/oauth2"

ROOT_URLCONF = "enterprise_data.urls"

SECRET_KEY = "insecure-secret-key"

USE_TZ = True
TIME_ZONE = 'UTC'

SITE_NAME = 'analytics-data-api'

ENTERPRISE_REPORTING_DB_ALIAS = 'default'
ENROLLMENTS_PAGE_SIZE = 10000

# Required for use with edx-drf-extensions JWT functionality:
# USER_SETTINGS overrides for djangorestframework-jwt APISettings class
# See https://github.com/GetBlimp/django-rest-framework-jwt/blob/master/rest_framework_jwt/settings.py
JWT_AUTH = {
    'JWT_AUDIENCE': 'test-aud',
    'JWT_DECODE_HANDLER': 'edx_rest_framework_extensions.auth.jwt.decoder.jwt_decode_handler',
    'JWT_ISSUER': 'test-iss',
    'JWT_LEEWAY': 1,
    'JWT_SECRET_KEY': 'test-key',
    'JWT_SUPPORTED_VERSION': '1.0.0',
    'JWT_VERIFY_AUDIENCE': False,
    'JWT_VERIFY_EXPIRATION': True,
    'JWT_AUTH_HEADER_PREFIX': 'JWT',

    # JWT_ISSUERS enables token decoding for multiple issuers (Note: This is not a native DRF-JWT field)
    # We use it to allow different values for the 'ISSUER' field, but keep the same SECRET_KEY and
    # AUDIENCE values across all issuers.
    'JWT_ISSUERS': [
        {
            'ISSUER': 'test-issuer-1',
            'SECRET_KEY': 'test-secret-key',
            'AUDIENCE': 'test-audience',
        },
        {
            'ISSUER': 'test-issuer-2',
            'SECRET_KEY': 'test-secret-key',
            'AUDIENCE': 'test-audience',
        }
    ],
}

SYSTEM_TO_FEATURE_ROLE_MAPPING = {
    SYSTEM_ENTERPRISE_ADMIN_ROLE: [ENTERPRISE_DATA_ADMIN_ROLE],
    SYSTEM_ENTERPRISE_OPERATOR_ROLE: [ENTERPRISE_DATA_ADMIN_ROLE],
}



"""
Client for connecting to AWS S3.
"""

import boto3


class S3Client:
    """
    Client for connecting to AWS S3.
    """

    def __init__(self):
        pass

    def get_enterprise_report(self, report_name, file_handler):
        """
        Download a report file from S3
        report_name <string>: name of the report on S3
        file_handler <file>: file object opened in binary mode
        """
        bucket_name = 'edx-enterprise-reporting'
        s3 = boto3.client('s3')
        s3.download_fileobj(bucket_name, report_name, file_handler)



"""
Client for connecting to a Snowflake database.
"""

import datetime
import os
from logging import getLogger

from snowflake import connector

LOGGER = getLogger(__name__)


class SnowflakeClient:
    """
    Client for connecting to Snowflake.
    """

    def __init__(self, username=None, password=None, account=None):
        """
        Instantiate a new client using the Django settings to determine the Snowflake credentials.
        If there are none configured, throw an exception.
        """
        username = username or os.environ['SNOWFLAKE_USERNAME']
        password = password or os.environ['SNOWFLAKE_PASSWORD']
        account = account or os.environ['SNOWFLAKE_ACCOUNT']

        self.connection_info = {
            'user': username,
            'password': password,
            'account': account,
        }

        self.connection = None
        self.cursor = None

    def connect(self):
        """
        Open a connection to Snowflake.
        """
        self.connection = connector.connect(**self.connection_info)
        self.cursor = self.connection.cursor()

    def close_connection(self):
        """
        Close the connection to Snowflake.
        """
        self.cursor.close()
        self.connection.close()
        self.cursor = None
        self.connection = None

    def stream_results(self, query):
        """
        Streams the results for a query using the current connection.
        """
        for row in self.cursor.execute(query):
            formatted_row = []
            for value in row:
                if isinstance(value, datetime.datetime):
                    formatted_row.append(value.strftime('%Y-%m-%d %H:%M:%S'))
                else:
                    formatted_row.append(value)
            yield formatted_row



"""
Filters for enterprise data views.
"""

from logging import getLogger

from rest_framework import filters

from django.conf import settings
from django.db.models import Q

from enterprise_data.clients import EnterpriseApiClient
from enterprise_data.constants import ANALYTICS_API_VERSION_0, ANALYTICS_API_VERSION_1, ANALYTICS_API_VERSION_ATTR
from enterprise_data.models import EnterpriseLearnerEnrollment

# Admittedly this is sort of hacky because the use of "|" with 2 Q objects
# forces the ORM to use a LEFT OUTER JOIN, which is needed to return a user
# that has multiple enrollments where at least one has consent_granted=True
CONSENT_TRUE_OR_NOENROLL_Q = Q(enrollments__consent_granted=True) | Q(enrollments__isnull=True)
ENROLLMENTS_CONSENT_TRUE_OR_NOENROLL_Q = Q(enrollments__is_consent_granted=True) | Q(enrollments__isnull=True)

LOGGER = getLogger(__name__)


class FiltersMixin:
    """
    Util mixin for enterprise_data filters.
    """

    def get_enterprise_customer(self, enterprise_uuid):
        """
        Return enterprise customer for `enterprise_uuid`.
        """
        enterprise_client = EnterpriseApiClient(
            settings.BACKEND_SERVICE_EDX_OAUTH2_PROVIDER_URL,
            settings.BACKEND_SERVICE_EDX_OAUTH2_KEY,
            settings.BACKEND_SERVICE_EDX_OAUTH2_SECRET,
        )
        return enterprise_client.get_enterprise_customer(enterprise_uuid)


class ConsentGrantedFilterBackend(filters.BaseFilterBackend, FiltersMixin):
    """
    Filter backend for any view that needs to filter results where consent has not been granted.

    This requires that `CONSENT_GRANTED_FILTER` be set in the view as a class variable, to identify
    the object's relationship to the consent_granted field.

    If the enterprise is configured for externally managed data sharing consent, all enrollments will be
    returned
    """

    def filter_queryset(self, request, queryset, view):
        """
        Filter a queryset for results where consent has been granted.
        """
        enterprise_uuid = view.kwargs['enterprise_id']
        enterprise_customer = self.get_enterprise_customer(enterprise_uuid)
        # if the enterprise is configured for "externally managed" data sharing consent,
        # ignore the consent_granted column.
        if enterprise_customer.get('enforce_data_sharing_consent') != 'externally_managed':
            filter_kwargs = {view.CONSENT_GRANTED_FILTER: True}
            queryset = queryset.filter(**filter_kwargs)
        return queryset


class AuditEnrollmentsFilterBackend(filters.BaseFilterBackend, FiltersMixin):
    """
    Filter backend to exclude enrollments where enrollment mode is `audit`.

    This requires that `ENROLLMENT_MODE_FILTER` be set in the view as a class
    variable, to identify the object's relationship to the
    `user_current_enrollment_mode` field.
    """

    def exclude_audit_enrollments(self, view):
        """
        Determine if audit enrollments should be excluded.
        """
        # this will be passed from admin-portal to avoid api call to lms
        audit_enrollments = view.request.query_params.get('audit_enrollments')
        if audit_enrollments:
            return audit_enrollments == 'false'

        enterprise_uuid = view.kwargs['enterprise_id']
        enterprise_customer = self.get_enterprise_customer(enterprise_uuid)
        return enterprise_customer.get('enable_audit_data_reporting') is False

    def filter_queryset(self, request, queryset, view):
        """
        Filter out queryset for results where enrollment mode is `audit`.
        """
        if self.exclude_audit_enrollments(view):
            enterprise_uuid = view.kwargs['enterprise_id']
            LOGGER.info(f'[AuditEnrollmentsFilterBackend] excluding audit enrollments for: {enterprise_uuid}')
            # Filter out enrollments that have audit mode and do not have a coupon code or an offer.
            filter_query = {
                view.ENROLLMENT_MODE_FILTER: 'audit',
                view.COUPON_CODE_FILTER: None,
                view.OFFER_FILTER: None
            }
            queryset = queryset.exclude(**filter_query)

        return queryset


class AuditUsersEnrollmentFilterBackend(filters.BaseFilterBackend, FiltersMixin):
    """
    Filter backend to filter Enterprise Users with 'audit' mode enrollments.
    """

    def filter_queryset(self, request, queryset, view):
        """
        Filter out queryset on the basis of its enrollment mode.

        If `enable_audit_data_reporting` is not enabled then it will exclude the Users with 'audit' mode enrollment.
        """
        enterprise_uuid = view.kwargs['enterprise_id']
        enterprise_customer = self.get_enterprise_customer(enterprise_uuid)

        enable_audit_data_reporting = enterprise_customer.get('enable_audit_data_reporting')

        version = getattr(view, ANALYTICS_API_VERSION_ATTR)

        if version == ANALYTICS_API_VERSION_0:
            queryset = queryset if enable_audit_data_reporting else queryset.filter(
                ~Q(enrollments__user_current_enrollment_mode='audit')
            )
        elif version == ANALYTICS_API_VERSION_1:
            if not enable_audit_data_reporting:
                audit_enrollments_enterprise_user_ids = EnterpriseLearnerEnrollment.objects.filter(
                    enterprise_customer_uuid=enterprise_uuid,
                    enterprise_user_id__isnull=False,
                    user_current_enrollment_mode="audit",
                ).values_list(
                    'enterprise_user_id',
                    flat=True
                )
                audit_enrollments_enterprise_user_ids = list(audit_enrollments_enterprise_user_ids)
                LOGGER.info(
                    "[ELV_ANALYTICS_API_V1] Enterprise: [%s], AuditDataReporting: [%s], AuditEnrollments: [%s]",
                    enterprise_uuid,
                    enable_audit_data_reporting,
                    len(audit_enrollments_enterprise_user_ids)
                )
                queryset = queryset.exclude(enterprise_user_id__in=audit_enrollments_enterprise_user_ids)

        return queryset



# Generated by Django 3.2.19 on 2023-06-15 07:05

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('enterprise_data', '0030_auto_20230609_1353'),
    ]

    operations = [
        migrations.AddField(
            model_name='enterpriselearnerenrollment',
            name='course_product_line',
            field=models.CharField(max_length=64, null=True),
        ),
        migrations.AddField(
            model_name='enterpriselearnerenrollment',
            name='is_subsidy',
            field=models.BooleanField(default=False),
        ),
    ]



# Generated by Django 3.2.19 on 2023-07-04 08:18

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('enterprise_data', '0031_auto_20230615_0705'),
    ]

    operations = [
        migrations.AddField(
            model_name='enterpriseoffer',
            name='amount_offer_spent_exec_ed',
            field=models.FloatField(null=True),
        ),
        migrations.AddField(
            model_name='enterpriseoffer',
            name='amount_offer_spent_ocm',
            field=models.FloatField(null=True),
        ),
    ]



#!/usr/bin/env python



import os
import re
import sys

from setuptools import setup


def get_version(*file_paths):
    """
    Extract the version string from the file at the given relative path fragments.
    """
    filename = os.path.join(os.path.dirname(__file__), *file_paths)
    version_file = open(filename).read()
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")

VERSION = get_version("enterprise_data", "__init__.py")

if sys.argv[-1] == "tag":
    print("Tagging the version on github:")
    os.system("git tag -a %s -m 'version %s'" % (VERSION, VERSION))
    os.system("git push --tags")
    sys.exit()


def load_requirements(*requirements_paths):
    """
    Load all requirements from the specified requirements files.

    Requirements will include any constraints from files specified
    with -c in the requirements files.
    Returns a list of requirement strings.
    """
    # UPDATED VIA SEMGREP - if you need to remove/modify this method remove this line and add a comment specifying why.

    # e.g. {"django": "Django", "confluent-kafka": "confluent_kafka[avro]"}
    by_canonical_name = {}

    def check_name_consistent(package):
        """
        Raise exception if package is named different ways.

        This ensures that packages are named consistently so we can match
        constraints to packages. It also ensures that if we require a package
        with extras we don't constrain it without mentioning the extras (since
        that too would interfere with matching constraints.)
        """
        canonical = package.lower().replace('_', '-').split('[')[0]
        seen_spelling = by_canonical_name.get(canonical)
        if seen_spelling is None:
            by_canonical_name[canonical] = package
        elif seen_spelling != package:
            raise Exception(
                f'Encountered both "{seen_spelling}" and "{package}" in requirements '
                'and constraints files; please use just one or the other.'
            )

    requirements = {}
    constraint_files = set()

    # groups "pkg<=x.y.z,..." into ("pkg", "<=x.y.z,...")
    re_package_name_base_chars = r"a-zA-Z0-9\-_."  # chars allowed in base package name
    # Two groups: name[maybe,extras], and optionally a constraint
    requirement_line_regex = re.compile(
        r"([%s]+(?:\[[%s,\s]+\])?)([<>=][^#\s]+)?"
        % (re_package_name_base_chars, re_package_name_base_chars)
    )

    def add_version_constraint_or_raise(current_line, current_requirements, add_if_not_present):
        regex_match = requirement_line_regex.match(current_line)
        if regex_match:
            package = regex_match.group(1)
            version_constraints = regex_match.group(2)
            check_name_consistent(package)
            existing_version_constraints = current_requirements.get(package, None)
            # It's fine to add constraints to an unconstrained package,
            # but raise an error if there are already constraints in place.
            if existing_version_constraints and existing_version_constraints != version_constraints:
                raise BaseException(f'Multiple constraint definitions found for {package}:'
                                    f' "{existing_version_constraints}" and "{version_constraints}".'
                                    f'Combine constraints into one location with {package}'
                                    f'{existing_version_constraints},{version_constraints}.')
            if add_if_not_present or package in current_requirements:
                current_requirements[package] = version_constraints

    # Read requirements from .in files and store the path to any
    # constraint files that are pulled in.
    for path in requirements_paths:
        with open(path) as reqs:
            for line in reqs:
                if is_requirement(line):
                    add_version_constraint_or_raise(line, requirements, True)
                if line and line.startswith('-c') and not line.startswith('-c http'):
                    constraint_files.add(os.path.dirname(path) + '/' + line.split('#')[0].replace('-c', '').strip())

    # process constraint files: add constraints to existing requirements
    for constraint_file in constraint_files:
        with open(constraint_file) as reader:
            for line in reader:
                if is_requirement(line):
                    add_version_constraint_or_raise(line, requirements, False)

    # process back into list of pkg><=constraints strings
    constrained_requirements = [f'{pkg}{version or ""}' for (pkg, version) in sorted(requirements.items())]
    return constrained_requirements


def is_requirement(line):
    """
    Return True if the requirement line is a package requirement.

    Returns:
        bool: True if the line is not blank, a comment,
        a URL, or an included file
    """
    # UPDATED VIA SEMGREP - if you need to remove/modify this method remove this line and add a comment specifying why

    return line and line.strip() and not line.startswith(('-r', '#', '-e', 'git+', '-c'))


setup(
    name="edx-enterprise-data",
    version=VERSION,
    classifiers=[
        'Framework :: Django',
        'Framework :: Django :: 3.2',
        'Framework :: Django :: 4.0',
        'Framework :: Django :: 4.2',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
    ],
    description="""Enterprise Reporting""",
    long_description="Tools and products related to providing access to Enterprise data.",
    author="edX",
    author_email="oscm@edx.org",
    url="https://github.com/openedx/edx-enterprise-data",
    packages=[
        'enterprise_data',
        'enterprise_reporting',
        'enterprise_data_roles',
    ],
    include_package_data=True,
    install_requires=load_requirements('requirements/base.in'),
    extras_require={'reporting':load_requirements('requirements/reporting.in')},
    license="AGPL 3.0",
    zip_safe=False,
)



"""
Clients used to connect to other systems.
"""


import logging
from urllib.parse import urljoin

from edx_django_utils.cache import TieredCache
from edx_rest_api_client.client import OAuthAPIClient
from requests.exceptions import HTTPError, RequestException
from rest_framework.exceptions import NotFound, ParseError

from django.conf import settings

from enterprise_data.utils import get_cache_key

DEFAULT_REPORTING_CACHE_TIMEOUT = 60 * 60 * 6  # 6 hours (Value is in seconds)
LOGGER = logging.getLogger('enterprise_data')


class EnterpriseApiClient(OAuthAPIClient):
    """
    The EnterpriseApiClient is used to communicate with the enterprise API endpoints in the LMS.

    This class is a sub-class of the OAuthAPIClient
    (https://github.com/openedx/edx-rest-api-client).
    """

    API_BASE_URL = urljoin(settings.LMS_BASE_URL + '/', 'enterprise/api/v1/')

    def get_enterprise_learner(self, user):
        """
        Get an enterprise learner record for a given user with the enterprise association.

        Returns: learner record or None if unable to retrieve or no Enterprise Learner exists
        """
        try:
            querystring = {'username': user.username}
            url = urljoin(self.API_BASE_URL, 'enterprise-learner')
            response = self.get(url, params=querystring)
            response.raise_for_status()
            data = response.json()
        except (HTTPError, RequestException) as exc:
            LOGGER.warning(
                "[Data Overview Failure] Unable to retrieve Enterprise Customer Learner details. "
                f"User: {user.username}, Exception: {exc}"
            )
            raise

        if data.get('results', None) is None:
            LOGGER.warning(
                f"[Data Overview Failure] Enterprise Customer Learner details could not be found. User: {user.username}"
            )
            raise NotFound('Unable to process Enterprise Customer Learner details for user {}: No Results Found'
                           .format(user.username))

        if data['count'] > 1:
            LOGGER.warning(
                f"[Data Overview Failure] Multiple Enterprise Customer Learners found. User: {user.username}"
            )
            raise ParseError(f'Multiple Enterprise Customer Learners found for user {user.username}')

        if data['count'] == 0:
            return None

        return data['results'][0]

    def get_enterprise_customer(self, enterprise_uuid):
        """
        Get the enterprises that this user has access to.
        """
        LOGGER.info(f'[EnterpriseApiClient] getting latest info for enterprise customer:{enterprise_uuid}')
        cache_key = get_cache_key(
            resource='enterprise-customer',
            enterprise_customer=enterprise_uuid,
        )
        cached_response = TieredCache.get_cached_response(cache_key)
        if cached_response.is_found:
            LOGGER.info(
                f'[EnterpriseApiClient] cache info found for enterprise customer:{enterprise_uuid}'
                f' with {cached_response.value}'
            )
            return cached_response.value

        LOGGER.info(f'[EnterpriseApiClient] No cached info found for enterprise customer:{enterprise_uuid}')
        url = urljoin(self.API_BASE_URL, f'enterprise-customer/{enterprise_uuid}')

        try:
            response = self.get(url)
            response.raise_for_status()
            data = response.json()
        except (HTTPError, RequestException) as exc:
            LOGGER.warning(
                "[Data Overview Failure] Unable to retrieve Enterprise Customer details. "
                f"Enterprise: {enterprise_uuid}, Exception: {exc}"
            )
            raise

        LOGGER.info(f'[EnterpriseApiClient] Setting cached for enterprise customer:{enterprise_uuid} with {data}')
        TieredCache.set_all_tiers(cache_key, data, DEFAULT_REPORTING_CACHE_TIMEOUT)

        return data
